open Phrase Environment
infix ++

type typeEnvironment = int environment

exception Error of string


fun verify(env,x,t) =
    let 
	val t' = get(env,x)
    in
	if t=t' then ()
	else raise Error("type error: " ^
			 "expected " ^ Int.toString t ^
			 ", got " ^ Int.toString t')
    end
    handle NotInEnvironment => raise Error("unknown identifier")

fun distinct(nil  ) = true
  | distinct(x::ys) = List.all (fn y => x<>y) ys andalso distinct(ys)

fun verifyParams(xs) = if distinct xs then ()
		       else raise Error("parameters not distinct")

fun exp(e,env) =
    case e of 
	CExp(i)       => ()
      | IExp(x)       => verify(env,x,0)
      | Cond(c,e1,e2) => (cond(c,env); exp(e1,env); exp(e2,env))
      | Let(ds,e)     => exp(e, decList(ds,env))
      | Appl(x, es)   => (  verify(env, x, List.length(es))
			  ; List.app (fn e => exp(e,env)) es
			  )

and cond(c, env) = (* condition * type environment
                      -> unit (* Error *)
		    *)


and dec(d,env) =
    case d of 
	Val(x, e)     => en[(x, 0)] before exp(e,env)
      | Fun(x, ys, e) => let 
			     val env'  = en[(x, List.length(ys))]
			     val env'' = en(List.map (fn y => (y,0)) ys)
			 in  
			     verifyParams(ys)
			   ; exp(e, env ++ env' ++ env'')
			   ; env'
			 end

and decList(ds, env) = (* declaration list * type environment
                          -> type environment (* Error *)
			*)
	    

fun elab(d, env) = env ++ dec(d,env)


val base = en[("~", 1), ("op+", 2), ("op-", 2), ("op*", 2)]
    
