open Phrase Environment
infix ++

datatype value =
    Int of int
  | DecProc  of declaration * valueEnvironment
  | PrimProc of value list -> value

withtype valueEnvironment = value environment

exception Error of string


fun toInt(Int i) = i

fun exp(e,env) =
    case e of 
	CExp(i)       => Int(i)
      | IExp(x)       => get(env,x)
      | Cond(c,e1,e2) => if cond(c,env) then exp(e1,env) else exp(e2,env)
      | Let(ds,e)     => exp(e, decList(ds,env))
      | Appl(x, es)   => apply(get(env,x), List.map (fn e => exp(e,env)) es)
	    
and apply(v, vs) =
    case v of
	PrimProc(p)                => p vs
      | DecProc(Fun(x,xs,e), env') => exp(e, env' ++ en[(x,v)] ++ 
					     en(ListPair.zip(xs,vs)))

and cond(c,env) =     (* condition * value environment
                         -> bool (* Overflow *)
		       *)

and dec(d,env) =      (* declaration * value environment
                         -> value environemnt (* Overflow *)
		       *)

and decList(ds,env) = (* declaration list * value environment
                         -> value environemnt (* Overflow *)
		       *)

fun eval(d,env) =
    let 
	val env' = dec(d,env) 
	           handle 
                     Overflow => raise Error "overflow"
		   |    _     => raise Error "wrong type"
	val x    = case d of Fun(x,_,_) => x | Val(x,_) => x
	val s    = case get(env',x) of Int(i) => Int.toString(i) | _ => "fn"
    in
	(x ^ " = " ^ s, env++env')
    end

val base = en[("~",   PrimProc(fn [Int(x)] => Int(~x))),
	      ("op+", PrimProc(fn [Int(x), Int(y)] => Int(x+y))),
	      ("op-", PrimProc(fn [Int(x), Int(y)] => Int(x-y))),
	      ("op*", PrimProc(fn [Int(x), Int(y)] => Int(x*y)))
	     ]
    
