open Token

exception Error of string

fun error cs = raise Error(String.implode cs)

val reservedSML = ["abstype", "and", "andalso", "as", "case",
		   "datatype", "do", "eqtype", "exception", "false",
		   "fn", "functor", "handle", "include", "infix",
		   "infixr", "local", "nil", "nonfix", "of", "op",
		   "open", "orelse", "raise", "rec", "ref", "sharing",
		   "sig", "signature", "struct", "structure", "true",
		   "type", "with", "while", "withtype", "where", "Bind",
		   "Match"]
	
fun lex'(cs, ts) =
    case cs of
	nil	        	   => rev(ts)
      | #" " :: cr                 => lex'(cr, ts)
      | #"\t" :: cr                => lex'(cr, ts)
      | #"\n" :: cr                => lex'(cr, ts)
      | #"(" :: cr		   => lex'(cr, LEFTP::ts)
      | #")" :: cr		   => lex'(cr, RIGHTP::ts)
      | #"," :: cr		   => lex'(cr, COMMA::ts)
      | #"=" :: #"~" :: _	   => error cs
      | #"<" :: #"=" :: #"~" :: _  => error cs
      | #"<" :: #"~" :: cr	   => error cs
      | #"=" :: cr		   => lex'(cr, EQUAL::ts)
      | #"<" :: #"=" :: cr	   => lex'(cr, LESSEQ::ts)
      | #"<" :: cr		   => lex'(cr, LESS::ts)
      | #"~" :: c :: cr	           => if Char.isDigit(c)
				      then constant(c::cr, ~1, ts)
				      else lex'(c::cr, IDENT("~")::ts)
      | #"o" :: #"p" :: #"+" :: cr => lex'(cr, IDENT("op+")::ts)
      | #"o" :: #"p" :: #"-" :: cr => lex'(cr, IDENT("op-")::ts)
      | #"o" :: #"p" :: #"*" :: cr => lex'(cr, IDENT("op*")::ts)
      | c :: cr		           => if Char.isDigit(c)
				      then constant(cs, 1, ts)
				      else if Char.isAlpha(c)
				           then idOrReserved(cr, [c], ts)
					   else error cs

and constant(cs, sign, ts) = 
    if cs <> nil andalso Char.isDigit(hd cs)
    then constant'(cs, 0, sign, ts)
	 handle Overflow => 
	     raise Error("integer constant too large: " ^ 
		         String.implode cs)				  
    else error cs

and constant'(cs, n, sign, ts) =
    if cs <> nil andalso Char.isDigit(hd cs)
    then constant'((tl cs),
		   10*n + Char.ord(hd cs) - Char.ord(#"0"),
		   sign,
		   ts)
    else lex'(cs, CONST(sign*n)::ts)

and idOrReserved(cs, xs, ts) =
    if cs <> nil andalso Char.isAlphaNum(hd cs)
    then idOrReserved(tl cs, hd cs :: xs, ts)
    else 
	case String.implode(List.rev(xs)) of 
	    "if"   => lex'(cs, IF::ts)
	  | "then" => lex'(cs, THEN::ts)
	  | "else" => lex'(cs, ELSE::ts)
	  | "let"  => lex'(cs, LET::ts)
	  | "in"   => lex'(cs, IN::ts)
	  | "end"  => lex'(cs, END::ts)
	  | "fun"  => lex'(cs, FUN::ts)
	  | "val"  => lex'(cs, VAL::ts)
	  | s      => if List.exists (fn x => x=s) reservedSML
		      then raise error(rev xs @ cs)
		      else lex'(cs, IDENT(s)::ts)

fun lex(cs) = lex'(cs,nil)




