open Token Phrase

exception Error of string

fun parse'(first, next, finished) =
    let
	fun match(t) = if t=next() then () 
		       else raise Error "wrong token"

	fun ident() =
	    case next() of 
		IDENT(s) => s
	      | _        => raise Error "identifier expected"

	fun tuple(arg) =
	    let 
		fun arglist() =
		    let 
			val x = arg()
		    in  
			if first()=COMMA then (next(); x::arglist())
			else [x]
		    end
		val _  = match(LEFTP)
		val xs = arglist()
		val _  = match(RIGHTP)
	    in  
		xs
	    end

	fun exp() =
	    case next() of 
		CONST i => CExp i
	      | IDENT s => if finished() orelse first() <> LEFTP
			   then IExp s
			   else Appl(s, tuple(exp))
	      | IF      => let 
			       val c  = cond()
			       val _  = match(THEN)
			       val e1 = exp()
			       val _  = match(ELSE)
			       val e2 = exp()
			   in  
			       Cond(c,e1,e2)
			   end
	      | LET      => let val ds  = declist()
				val _   = match(IN)
				val e   = exp()
				val _   = match(END)
			    in  
				Let(ds,e)
			    end
	      | _       => raise Error "expression expected"

	and cond() =
	    let 
		val e1 = exp()
		val c  = (case next() of
			      EQUAL  => Equal
			    | LESS   => Less
			    | LESSEQ => LessEq
			    | _      => raise Error "condition expected")
		val e2 = exp()
	    in  
		c(e1,e2)
	    end		      

	and dec() =
	    case first() of 
		FUN => let val _  = next()
			   val x  = ident()
			   val xs = tuple(ident)
			   val _  = match(EQUAL)
			   val e  = exp()
		       in  
			   Fun(x,xs,e)
		       end
	      | VAL => let 
			   val _  = next()
			   val x  = ident()
			   val _  = match(EQUAL)
			   val e  = exp()
		       in  
			   Val(x,e)
		       end
	      | _   => Val("it", exp())
		      
	and declist() =
	    if first()=FUN orelse first()=VAL
	    then dec() :: declist()
	    else nil
    in
	dec() 
	before 
	(if finished() then () else raise Error "eof expected")
    end

fun parse(ts) =
    let
	val r          = ref ts
	fun first()    = (case !r of 
			      t::_ => t 
			    | _    => raise Error "unexpected eof")
	fun next()     = first() before r := tl(!r)
	fun finished() = !r = nil
    in
	parse'(first, next, finished)
    end
