val elabOn = ref true (* elaboration can be switched off *)

fun interpretSentence'(cs, env as (typeEnv, valueEnv)) =
    let
	val dec            = Parse.parse(Lex.lex cs)
	val typeEnv'       = if !elabOn then Elab.elab(dec, typeEnv)
			     else typeEnv
	val (s, valueEnv') = Eval.eval(dec, valueEnv)
    in
	(typeEnv', valueEnv')
	before
	print("> " ^ s ^ "\n")
    end

fun interpretSentence(cs, env) =
    let
	fun continue(s,s') = 
	    env before print("! " ^ s ^ " Error: " ^ s' ^ "\n")
    in
	interpretSentence'(cs, env)
	handle
	  Lex.Error(s)   => continue("Lexical", s)
	| Parse.Error(s) => continue("Syntax", s)
	| Elab.Error(s)  => continue("Elaboration" , s)
	| Eval.Error(s)  => continue("Evaluation" , s)
    end
	  
fun interpret'(nextSentence, env) =
    case nextSentence() of
	nil => env
      | cs  => interpret'(nextSentence, interpretSentence(cs,env))

fun interpret(nextString, env) =
    let
	val rest = ref nil

	fun nextSentence(xs, ys) =
	    case xs of 
		nil      => (case String.explode(nextString()) of 
				 nil => nil
			       | xs  => nextSentence(xs,ys))
	      | #";"::xr => rev(ys) before rest:=xr
	      |    x::xr => nextSentence(xr, x::ys)
    in
	interpret'(fn () => nextSentence(!rest, nil), env)
    end

fun read() = TextIO.inputLine(TextIO.stdIn)

fun readWord() = 
    let val s = read() in String.substring(s, 0, size s - 1) end

fun readMML() = (print("- "); read())

fun topLevel(env) =
    (  print("(quit, use, elab off, continue)? ")
     ; case read() of 
          ""           => print "\n"
	| "quit\n"     => print "\n"
	| "continue\n" => topLevel(interpret(readMML,env))
	| "use\n"      => topLevel(use(env))
	| "elab off\n" => (  print "elaboration switched  off\n"
                           ; elabOn := false
			   ; topLevel(env)
			   )
	| _            => (  print("! ignored\n")
			   ; topLevel(env)
			   )
     )

and use(env) =
    let
	val _        = print("file name: ")
	val inStream = TextIO.openIn(readWord())
	fun read()   = TextIO.inputLine(inStream)
    in
	interpret(read, env)
	before
	TextIO.closeIn(inStream)
    end
    handle 
      _ => env 
	   before 
           print("! file not readable\n")


val _ = topLevel(interpret(readMML, (Elab.base, Eval.base)))
        handle _ => print "interpreter error\n"

    
